import mchammer as mch  # noqa: INP001, D100
import spindry as spd
import stk

# Building a cage from the examples on the stk docs.
bb1 = stk.BuildingBlock(
    smiles="O=CC(C=O)C=O",
    functional_groups=[stk.AldehydeFactory()],
)
bb2 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])

cage = stk.ConstructedMolecule(
    topology_graph=stk.cage.FourPlusSix(
        building_blocks=(bb1, bb2),
    ),
)
cage.write("poc.mol")
# Always ensure initial structures are not entirely on top of each
# other.
stk_guests = (
    (stk.BuildingBlock("CN1C=NC2=C1C(=O)N(C(=O)N2C)C"), (0.0, 0.0, 0.0)),
    (stk.BuildingBlock("c1ccccc1"), (0.0, 2.0, 0.0)),
)

host_guest = stk.ConstructedMolecule(
    topology_graph=stk.host_guest.Complex(
        host=stk.BuildingBlock.init_from_molecule(cage),
        guests=[
            stk.host_guest.Guest(i[0], displacement=i[1]) for i in stk_guests
        ],
    )
)
host_guest.write("host_multi_guest.mol")

supramolecule = spd.SupraMolecule(
    atoms=(
        mch.Atom(
            id=atom.get_id(),
            element_string=atom.__class__.__name__,
        )
        for atom in host_guest.get_atoms()
    ),
    bonds=(
        mch.Bond(
            id=i,
            atom_ids=(
                bond.get_atom1().get_id(),
                bond.get_atom2().get_id(),
            ),
        )
        for i, bond in enumerate(host_guest.get_bonds())
    ),
    position_matrix=host_guest.get_position_matrix(),
)
print(supramolecule)  # noqa: T201

cg = spd.Spinner(
    step_size=1,
    rotation_step_size=5,
    num_conformers=10,
)

conformer = cg.get_final_conformer(supramolecule)
print(conformer)  # noqa: T201
print(conformer.get_cid(), conformer.get_potential())  # noqa: T201

# Write optimised structure out.
opt_host_guest = host_guest.with_position_matrix(
    conformer.get_position_matrix()
)
opt_host_guest.write("host_multi_guest_out.mol")



"""This module defines general-purpose objects, functions and classes."""

import itertools as it

import numpy as np
from scipy.spatial.distance import cdist, euclidean

from .supramolecule import SupraMolecule


def get_atom_distance(
    position_matrix: np.ndarray,
    atom1_id: int,
    atom2_id: int,
) -> float:
    """Return the distance between two atoms."""
    return float(
        euclidean(u=position_matrix[atom1_id], v=position_matrix[atom2_id])
    )


def calculate_min_atom_distance(supramolecule: SupraMolecule) -> float:
    """Calculate the minimum distance between components in supramolecule."""
    component_position_matrices = (
        i.get_position_matrix() for i in supramolecule.get_components()
    )

    min_distance = 1e24
    for pos_mat_pair in it.combinations(component_position_matrices, 2):
        pair_dists = cdist(pos_mat_pair[0], pos_mat_pair[1])
        min_distance = min([min_distance, min(pair_dists.flatten())])

    return min_distance


def calculate_centroid_distance(supramolecule: SupraMolecule) -> float:
    """Calculate the centroid distances in 1:1 complex."""
    comps = list(supramolecule.get_components())
    if len(comps) != 2:  # noqa: PLR2004
        msg = "more than one guest there buddy!"
        raise ValueError(msg)

    return float(
        np.linalg.norm(comps[0].get_centroid() - comps[1].get_centroid())
    )



import mchammer as mch
import numpy as np
import spindry as spd

from tests.utilities import is_equivalent_spd_molecule


def test_opt(
    spinner: spd.Spinner,
    smolecule: mch.Molecule,
    final_pos_mat: np.ndarray,
    final_potential: float,
) -> None:
    test = spinner.get_final_conformer(smolecule)
    print(test.get_position_matrix(), final_pos_mat)
    assert np.all(
        np.allclose(
            final_pos_mat,
            test.get_position_matrix(),
        )
    )
    print(spinner.compute_potential(test), final_potential)
    assert np.isclose(spinner.compute_potential(test), final_potential)


def test_opt_spd(
    spd_spinner: spd.Spinner,
    spd_host: mch.Molecule,
    spd_guest: mch.Molecule,
    spd_supramolecule: spd.SupraMolecule,
) -> None:
    test = spd_spinner.get_final_conformer(spd_supramolecule)
    is_equivalent_spd_molecule(test, spd_supramolecule)
    for i, comp in enumerate(list(test.get_components())):
        if i == 0:
            test_host = spd_host.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_spd_molecule(test_host, spd_host)
        elif i == 1:
            test_guest = spd_guest.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_spd_molecule(test_guest, spd_guest)


def test_opt_spd_components(
    spd_spinner: spd.Spinner,
    spd_host: mch.Molecule,
    spd_guest: mch.Molecule,
    spd_supramolecule: spd.SupraMolecule,
) -> None:
    test = spd_spinner.get_final_conformer(spd_supramolecule)
    is_equivalent_spd_molecule(test, spd_supramolecule)
    for i, comp in enumerate(list(test.get_components())):
        if i == 0:
            test_host = spd_host.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_spd_molecule(test_host, spd_host)
        elif i == 1:
            test_guest = spd_guest.with_position_matrix(
                comp.get_position_matrix()
            )
            is_equivalent_spd_molecule(test_guest, spd_guest)


def test_opt_setcomp1(
    spinner: spd.Spinner,
    smolecule_components: spd.SupraMolecule,
    final_comp_pos_mat1: np.ndarray,
    movable_components1: tuple,
) -> None:
    test = spinner.get_final_conformer(
        smolecule_components,
        movable_components1,
    )
    print(test.get_position_matrix(), final_comp_pos_mat1)
    assert np.all(
        np.allclose(
            final_comp_pos_mat1,
            test.get_position_matrix(),
        )
    )


def test_opt_setcomp2(
    spinner: spd.Spinner,
    smolecule_components: spd.SupraMolecule,
    final_comp_pos_mat2: np.ndarray,
    movable_components2: tuple,
) -> None:
    test = spinner.get_final_conformer(
        smolecule_components,
        movable_components2,
    )
    print(test.get_position_matrix(), final_comp_pos_mat2)
    assert np.all(
        np.allclose(
            final_comp_pos_mat2,
            test.get_position_matrix(),
        )
    )



"""Generator of host guest conformations using nonbonded interactions."""

from __future__ import annotations

from typing import TYPE_CHECKING

import mchammer as mch
import numpy as np

from .potential import SpdPotential
from .supramolecule import SupraMolecule

if TYPE_CHECKING:
    from collections import abc

    from .potential import Potential


class Spinner:
    """Generate host-guest conformations by rotating guest.

    A Metroplis MC algorithm is applied to perform rigid
    translations and rotations of the guest, relative to the host.

    """

    def __init__(  # noqa: PLR0913
        self,
        step_size: float,
        rotation_step_size: float,
        num_conformers: int,
        max_attempts: int = 1000,
        potential_function: Potential | None = None,
        beta: float = 2,
        random_seed: int | None = 1000,
    ) -> None:
        """Initialize a :class:`Spinner` instance.

        Parameters:
            step_size:
                The relative size of the step to take during step.

            rotation_step_size:
                The relative size of the rotation to take during step.

            num_conformers:
                Number of conformers to extract.

            max_attempts:
                Maximum number of MC moves to try to generate conformers.

            potential_function:
                Function to calculate potential energy of a
                :class:`spd.Supramolecule`

            beta:
                Value of beta used in the in MC moves. Beta takes the
                place of the inverse boltzmann temperature.
                Defaults to 2.

            random_seed:
                Random seed to use for MC algorithm. Should only be set to
                ``None`` if system-based random seed is desired. Defaults
                to a set seed of 1000, to avoid randomness.

        """
        self._step_size = step_size
        self._num_conformers = num_conformers
        self._rotation_step_size = rotation_step_size
        self._max_attempts = max_attempts
        if potential_function is None:
            self._potential_function = SpdPotential(5)
        else:
            self._potential_function = potential_function  # type: ignore[assignment]
        self._beta = beta
        if random_seed is None:
            self._generator = np.random.default_rng()
        else:
            self._generator = np.random.default_rng(random_seed)

    def compute_potential(self, supramolecule: SupraMolecule) -> float:
        """Compute the potential of a supramolecule."""
        return self._potential_function.compute_potential(supramolecule)

    def _run_step(
        self,
        supramolecule: SupraMolecule,
        movable_components: tuple[int, ...] | None,
    ) -> tuple[SupraMolecule, float]:
        component_list = list(supramolecule.get_components())
        component_sizes = {
            i: mol.get_num_atoms() for i, mol in enumerate(component_list)
        }
        max_size = max(component_sizes.values())

        # If movable components not set, select a guest randomly to
        # move and reorient.
        # Do not move or rotate largest component.
        if movable_components is None:
            # If there are different sizes.
            if len(set(component_sizes.values())) > 1:
                movable_components = tuple(
                    i
                    for i in range(len(component_list))
                    if component_sizes[i] != max_size
                )
            # Else capture all!
            else:
                movable_components = tuple(
                    i for i in range(len(component_list))
                )

        targ_comp_id = self._generator.choice(
            [i for i in range(len(component_list)) if i in movable_components]
        )

        targ_comp = component_list[targ_comp_id]

        # Random number from -1 to 1 for multiplying translation.
        rand = (self._generator.random() - 0.5) * 2

        # Random translation direction.
        rand_vector = self._generator.random(3)
        rand_vector = rand_vector / np.linalg.norm(rand_vector)

        # Perform translation.
        translation_vector = rand_vector * self._step_size * rand
        targ_comp = mch.translate_molecule_along_vector(
            mol=targ_comp,
            vector=translation_vector,
        )

        # Define a random rotation of the guest.
        # Random number from -1 to 1 for multiplying rotation.
        rand = (self._generator.random() - 0.5) * 2
        rotation_angle = self._rotation_step_size * rand
        rand_axis = self._generator.random(3)
        rand_axis = rand_axis / np.linalg.norm(rand_vector)

        # Perform rotation.
        targ_comp = mch.rotate_molecule_by_angle(
            mol=targ_comp,
            angle=rotation_angle,
            axis=rand_axis,
            origin=targ_comp.get_centroid(),
        )

        component_list[targ_comp_id] = targ_comp
        supramolecule = SupraMolecule.init_from_components(
            components=component_list,
        )

        nonbonded_potential = self.compute_potential(supramolecule)
        return supramolecule, nonbonded_potential

    def get_conformers(
        self,
        supramolecule: SupraMolecule,
        movable_components: tuple[int, ...] | None = None,
        verbose: bool = False,  # noqa: FBT001, FBT002
    ) -> abc.Iterable[mch.Molecule]:
        """Get conformers of supramolecule.

        Parameters:
            supramolecule:
                The supramolecule to optimize.

            movable_components:
                Components of supramolecule to move during simulation.

            verbose:
                `True` to print some extra information.

        Yields:
            conformer: :class:`.SupraMolecule`
                The host-guest supramolecule.

        """
        cid = 0
        nonbonded_potential = self.compute_potential(supramolecule)

        yield SupraMolecule.init_from_components(
            components=list(supramolecule.get_components()),
            cid=cid,
            potential=nonbonded_potential,
        )
        cids_passed = []
        count = 0
        for _ in range(1, self._max_attempts):
            n_supramolecule, n_nonbonded_potential = self._run_step(
                supramolecule=supramolecule,
                movable_components=movable_components,
            )
            passed = mch.test_move(
                beta=self._beta,
                curr_pot=nonbonded_potential,
                new_pot=n_nonbonded_potential,
                generator=self._generator,
            )
            if passed:
                cid += 1
                cids_passed.append(cid)
                nonbonded_potential = n_nonbonded_potential
                supramolecule = SupraMolecule.init_from_components(
                    components=list(n_supramolecule.get_components()),
                    cid=cid,
                    potential=nonbonded_potential,
                )
                supramolecule = supramolecule.with_position_matrix(
                    n_supramolecule.get_position_matrix()
                )
                yield supramolecule
            count += 1
            if len(cids_passed) == self._num_conformers:
                break

        if verbose:
            print(  # noqa: T201
                f"{len(cids_passed)} conformers generated in {count} steps."
            )

    def get_final_conformer(
        self,
        supramolecule: SupraMolecule,
        movable_components: tuple[int, ...] | None = None,
    ) -> mch.Molecule:
        """Get final conformer of supramolecule.

        Parameters:
            supramolecule:
                The supramolecule to optimize.

            movable_components:
                Components of supramolecule to move during simulation.
                If `None`, then moved components are selected randomly,
                and the largest component (host) is not moved.

        Returns:
            conformer:
                The host-guest supramolecule.

        """
        for conformer in self.get_conformers(  # noqa: B007
            supramolecule=supramolecule,
            movable_components=movable_components,
        ):
            continue

        return conformer



import numpy as np
import spindry as spd


def test_smolecule_define_components(
    smolecule: spd.SupraMolecule,
    components: list,
) -> None:
    tests = list(smolecule.get_components())

    assert len(tests) == len(components)
    for test, comp in zip(tests, components):
        for a1, a2 in zip(test.get_atoms(), comp.get_atoms()):
            assert a1.get_id() == a2.get_id()
            assert a1.get_element_string() == a2.get_element_string()

        assert np.all(
            np.allclose(
                test.get_position_matrix(),
                comp.get_position_matrix(),
            )
        )



import itertools as it

import mchammer as mch


def is_equivalent_spd_atom(atom1: mch.Atom, atom2: mch.Atom) -> None:
    """Test if atoms are equivalent."""
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_element_string() == atom2.get_element_string()
    assert atom1.get_radius() == atom2.get_radius()
    assert atom1.__class__ is atom2.__class__


def is_equivalent_spd_bond(bond1: mch.Bond, bond2: mch.Bond) -> None:
    """Test if bonds are equivalent."""
    assert bond1.__class__ is bond2.__class__
    assert bond1.get_id() == bond2.get_id()
    assert bond1.get_atom1_id() == bond2.get_atom1_id()
    assert bond1.get_atom2_id() == bond2.get_atom2_id()


def is_equivalent_spd_molecule(
    molecule1: mch.Molecule,
    molecule2: mch.Molecule,
) -> None:
    """Test if molecules are equivalent."""
    print(molecule1, molecule2)
    atoms = it.zip_longest(
        molecule1.get_atoms(),
        molecule2.get_atoms(),
    )
    for atom1, atom2 in atoms:
        is_equivalent_spd_atom(atom1, atom2)

    bonds = it.zip_longest(
        molecule1.get_bonds(),
        molecule2.get_bonds(),
    )
    for bond1, bond2 in bonds:
        is_equivalent_spd_bond(bond1, bond2)



import stk

bb1 = stk.BuildingBlock(
    smiles="NCCN",
    functional_groups=[stk.PrimaryAminoFactory()],
)
bb2 = stk.BuildingBlock(
    smiles="O=CC(C=O)C=O",
    functional_groups=[stk.AldehydeFactory()],
)
guest1 = stk.host_guest.Guest(
    building_block=stk.BuildingBlock("c1ccccc1"),
)
guest2 = stk.host_guest.Guest(
    building_block=stk.BuildingBlock("C1CCCCC1"),
)
cage = stk.ConstructedMolecule(
    topology_graph=stk.cage.FourPlusSix(
        building_blocks=(bb1, bb2),
        optimizer=stk.MCHammer(),
    ),
)

hgcomplex = stk.ConstructedMolecule(
    topology_graph=stk.host_guest.Complex(
        host=stk.BuildingBlock.init_from_molecule(cage),
        guests=(guest1, guest2),
        optimizer=stk.Spinner(),
    ),
)
hgcomplex.write("spinner.mol")



import pathlib

import numpy as np
import spindry as spd
import stk


def test_stk_spinner() -> None:
    bb1 = stk.BuildingBlock(
        smiles="NCCN",
        functional_groups=[stk.PrimaryAminoFactory()],
    )
    bb2 = stk.BuildingBlock(
        smiles="O=CC(C=O)C=O",
        functional_groups=[stk.AldehydeFactory()],
    )
    guest1 = stk.host_guest.Guest(
        building_block=stk.BuildingBlock("c1ccccc1"),
    )
    guest2 = stk.host_guest.Guest(
        building_block=stk.BuildingBlock("C1CCCCC1"),
    )
    cage = stk.ConstructedMolecule(
        topology_graph=stk.cage.FourPlusSix(
            building_blocks=(bb1, bb2),
            optimizer=stk.MCHammer(),
        ),
    )

    complex_opt = stk.ConstructedMolecule(
        topology_graph=stk.host_guest.Complex(
            host=stk.BuildingBlock.init_from_molecule(cage),
            guests=(guest1, guest2),
            optimizer=stk.Spinner(),
        ),
    )

    complex_unopt = stk.ConstructedMolecule(
        topology_graph=stk.host_guest.Complex(
            host=stk.BuildingBlock.init_from_molecule(cage),
            guests=(guest1, guest2),
        ),
    )
    supramolecule = spd.SupraMolecule(
        atoms=(
            spd.Atom(
                id=atom.get_id(),
                element_string=atom.__class__.__name__,
            )
            for atom in complex_unopt.get_atoms()
        ),
        bonds=(
            spd.Bond(
                id=i,
                atom_ids=(
                    bond.get_atom1().get_id(),
                    bond.get_atom2().get_id(),
                ),
            )
            for i, bond in enumerate(complex_unopt.get_bonds())
        ),
        position_matrix=complex_unopt.get_position_matrix(),
    )
    # Run optimization.
    optimizer = spd.Spinner(
        potential_function=spd.SpdPotential(
            nonbond_epsilon=5.0,
        ),
        step_size=1.5,
        rotation_step_size=5.0,
        num_conformers=50,
        max_attempts=1000,
        beta=2.0,
        random_seed=1000,
    )
    conformer = optimizer.get_final_conformer(supramolecule)
    new_complex = complex_unopt.with_position_matrix(
        position_matrix=conformer.get_position_matrix(),
    )

    assert np.all(
        np.equal(
            complex_opt.get_position_matrix(),
            new_complex.get_position_matrix(),
        )
    )

    known = stk.BuildingBlock.init_from_file(
        pathlib.Path(__file__).resolve().parent / "spinner.mol"
    )

    assert np.all(
        np.equal(
            known.get_position_matrix(),
            new_complex.get_position_matrix(),
        )
    )



import mchammer as mch  # noqa: INP001, D100
import numpy as np
import spindry as spd
import stk

# Building a cage from the examples on the stk docs.
bb1 = stk.BuildingBlock(
    smiles="O=CC(C=O)C=O",
    functional_groups=[stk.AldehydeFactory()],
)
bb2 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])

cage = stk.ConstructedMolecule(
    topology_graph=stk.cage.FourPlusSix(
        building_blocks=(bb1, bb2),
    ),
)
cage.write("poc.mol")
# Always ensure initial structures are not entirely on top of each
# other.
stk_guests = (
    (stk.BuildingBlock("CN1C=NC2=C1C(=O)N(C(=O)N2C)C"), (0.0, 0.0, 0.0)),
)

host_guest = stk.ConstructedMolecule(
    topology_graph=stk.host_guest.Complex(
        host=stk.BuildingBlock.init_from_molecule(cage),
        guests=[
            stk.host_guest.Guest(i[0], displacement=i[1]) for i in stk_guests
        ],
    )
)

supramolecule = spd.SupraMolecule(
    atoms=(
        mch.Atom(
            id=atom.get_id(),
            element_string=atom.__class__.__name__,
        )
        for atom in host_guest.get_atoms()
    ),
    bonds=(
        mch.Bond(
            id=i,
            atom_ids=(
                bond.get_atom1().get_id(),
                bond.get_atom2().get_id(),
            ),
        )
        for i, bond in enumerate(host_guest.get_bonds())
    ),
    position_matrix=host_guest.get_position_matrix(),
)
print(supramolecule)  # noqa: T201


class PotFn(spd.SpdPotential):
    """Scale the size of the guest radii."""

    def __init__(self, guest_scale: float, nonbond_epsilon: float = 5) -> None:
        """Initialize potential class."""
        self._guest_scale = guest_scale
        super().__init__(nonbond_epsilon)

    def compute_potential(self, supramolecule: spd.SupraMolecule) -> float:
        """Compute the potential."""
        component_position_matrices = (
            i.get_position_matrix() for i in supramolecule.get_components()
        )
        component_radii = [
            tuple(j.get_radius() for j in i.get_atoms())
            for i in supramolecule.get_components()
        ]

        component_radii[1] = [
            i * self._guest_scale for i in component_radii[1]
        ]
        return self._compute_nonbonded_potential(
            position_matrices=component_position_matrices,
            radii=component_radii,
        )


for i in np.arange(1, 2.1, 0.1):
    # Do not want to move, just get energy.
    cg = spd.Spinner(
        step_size=0,
        rotation_step_size=0,
        num_conformers=1,
        max_attempts=1,
        potential_function=PotFn(guest_scale=i),
    )

    conformer = cg.get_final_conformer(supramolecule)
    print(round(i, 2), round(conformer.get_potential(), 3))  # noqa: T201


class CentroidFn:
    """A potential function based on the minimum host-guest distance.

    Here we show that as long as you provide a `compute_potential` method, then
    any class works!

    """

    def compute_potential(self, supramolecule: spd.SupraMolecule) -> float:
        """Compute the potential."""
        centroids = [i.get_centroid() for i in supramolecule.get_components()]

        return 10 / np.linalg.norm(centroids[1] - centroids[0])


# Do not want to move, just get energy.
cg = spd.Spinner(
    step_size=1,
    rotation_step_size=2,
    num_conformers=100,
    max_attempts=1000,
    potential_function=CentroidFn(),
)
conformer = cg.get_final_conformer(supramolecule)
# Write optimised structure out.
opt_host_guest = host_guest.with_position_matrix(
    conformer.get_position_matrix()
)
opt_host_guest.write("custom_pot_fn_out.mol")



"""SupraMolecule class for optimisation."""

from __future__ import annotations

import typing
from dataclasses import dataclass

import mchammer as mch
import networkx as nx
import numpy as np

if typing.TYPE_CHECKING:
    from collections import abc


@dataclass
class SupraMolecule(mch.Molecule):
    """Representation of a supramolecule containing atoms and positions.

    Parameters:
        atoms : :class:`iterable` of :class:`.Atom`
            Atoms that define the molecule.

        bonds : :class:`iterable` of :class:`.Bond`
            Bonds between atoms that define the molecule.

        position_matrix : :class:`numpy.ndarray`
            A ``(n, 3)`` matrix holding the position of every atom in
            the :class:`.Molecule`.

        cid : :class:`int`, optional
            Conformer id of supramolecule.

        potential : :class:`float`, optional
            Potential energy of Supramolecule.

    """

    atoms: tuple[mch.Atom, ...]
    bonds: tuple[mch.Bond, ...]
    position_matrix: np.ndarray
    cid: int | None = None
    potential: float | None = None

    def __post_init__(self) -> None:
        """Post initialization of molecule."""
        self.atoms = tuple(self.atoms)
        self.bonds = tuple(self.bonds)
        self.position_matrix = np.array(
            self.position_matrix.T,
            dtype=np.float64,
        )

        self._define_components()

    def with_position_matrix(
        self,
        position_matrix: np.ndarray,
    ) -> SupraMolecule:
        """Return clone SupraMolecule with new position matrix.

        Parameters:
            position_matrix:
                A position matrix of the clone. The shape of the matrix
                is ``(n, 3)``.

        """
        _temp_components = tuple(self.get_components())

        _temp_supramolecule = SupraMolecule(
            atoms=tuple(self.atoms),
            bonds=tuple(self.bonds),
            position_matrix=np.array(position_matrix),
            cid=self.cid,
            potential=self.potential,
        )
        # Overwrite redefined components.
        _temp_supramolecule.components = _temp_components
        return _temp_supramolecule

    def with_displacement(self, displacement: np.ndarray) -> SupraMolecule:
        """Return a displaced clone Molecule.

        Parameters:
            displacement:
                The displacement vector to be applied.

        """
        new_position_matrix = self.position_matrix.T + displacement
        return SupraMolecule(
            atoms=tuple(self.atoms),
            bonds=tuple(self.bonds),
            cid=self.cid,
            potential=self.potential,
            position_matrix=np.array(new_position_matrix),
        )

    @classmethod
    def init_from_components(
        cls,  # noqa: ANN102
        components: list[mch.Molecule],
        cid: int | None = None,
        potential: float | None = None,
    ) -> typing.Self:
        """Initialize a :class:`Supramolecule` instance from components.

        Parameters:
            components:
                Molecular components that define the supramolecule.

            cid:
                Conformer id of supramolecule.

            potential:
                Potential energy of Supramolecule.

        """
        atoms = []
        bonds = []
        position_matrix = []
        # Map old atom ids in components to atom ids in supramolecule.
        atom_id_map: dict[int, int] = {}
        bond_id_map: dict[int, int] = {}
        for comp in components:
            for a in comp.get_atoms():
                if len(atom_id_map) == 0:
                    atom_id_map[a.get_id()] = 0
                else:
                    atom_id_map[a.get_id()] = (
                        max(list(atom_id_map.values())) + 1
                    )
                atoms.append(
                    mch.Atom(
                        id=atom_id_map[a.get_id()],
                        element_string=a.get_element_string(),
                    )
                )
            for b in comp.get_bonds():
                if len(bond_id_map) == 0:
                    bond_id_map[b.get_id()] = 0
                else:
                    bond_id_map[b.get_id()] = (
                        max(list(bond_id_map.values())) + 1
                    )
                bonds.append(
                    mch.Bond(
                        id=bond_id_map[b.get_id()],
                        atom_ids=(
                            atom_id_map[b.get_atom1_id()],
                            atom_id_map[b.get_atom2_id()],
                        ),
                    )
                )
            for pos in comp.get_position_matrix():
                position_matrix.append(pos)  # noqa: PERF402

        supramolecule: SupraMolecule = cls.__new__(cls)
        supramolecule.atoms = tuple(atoms)
        supramolecule.bonds = tuple(bonds)
        supramolecule.components = tuple(components)
        supramolecule.cid = cid
        supramolecule.potential = potential
        supramolecule.position_matrix = np.array(position_matrix).T
        return supramolecule

    def _define_components(self) -> None:
        """Define disconnected component molecules as :class:`.Molecule`s."""
        # Produce a graph from the molecule that does not include edges
        # where the bonds to be optimized are.
        mol_graph = nx.Graph()
        for atom in self.get_atoms():
            mol_graph.add_node(atom.get_id())

        # Add edges.
        for bond in self.bonds:
            pair_ids = (bond.get_atom1_id(), bond.get_atom2_id())
            mol_graph.add_edge(*pair_ids)

        # Get atom ids in disconnected subgraphs.
        comps = []
        for c in nx.connected_components(mol_graph):
            c_ids = sorted(c)
            in_atoms = [i for i in self.atoms if i.get_id() in c]
            in_bonds = [
                i
                for i in self.bonds
                if i.get_atom1_id() in c and i.get_atom2_id() in c
            ]
            new_pos_matrix = self.position_matrix[:, list(c_ids)].T
            comps.append(mch.Molecule(in_atoms, in_bonds, new_pos_matrix))

        self.components = tuple(comps)

    def write_xyz_content(self) -> list[str]:
        """Write basic `.xyz` file content of Molecule."""
        coords = self.get_position_matrix()
        content = ["0"]
        for i, atom in enumerate(self.get_atoms(), 1):
            x, y, z = (i for i in coords[atom.get_id()])
            content.append(f"{atom.get_element_string()} {x:f} {y:f} {z:f}\n")
        # Set first line to the atom_count.
        content[0] = f"{i}\ncid:{self.cid}, pot: {self.potential}\n"

        return content

    def get_components(self) -> abc.Iterable[mch.Molecule]:
        """Yields each molecular component."""
        yield from self.components

    def get_cid(self) -> int | None:
        """Get conformer id."""
        return self.cid

    def get_potential(self) -> float | None:
        """Get potential energy."""
        return self.potential

    def __str__(self) -> str:
        """String representation of SupraMolecule."""
        return repr(self)

    def __repr__(self) -> str:
        """String representation of SupraMolecule."""
        comps = ", ".join([str(i) for i in self.get_components()])
        return (
            f"{self.__class__.__name__}("
            f"{len(list(self.get_components()))} components, "
            f"{comps})"
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import spindry as spd


def test_nonbond_potential(
    spdpotential: spd.SpdPotential,
    distances: np.ndarray,
    nonbond_potentials: list[float],
    nb_mins: list[int],
) -> None:
    for i, d in enumerate(distances):
        test = spdpotential._nonbond_potential(  # noqa: SLF001
            distance=d,
            sigmas=np.array(1.2),
        )
        assert np.isclose(test, nonbond_potentials[i], atol=1e-5)

    for i, _sigma in enumerate([1.2, 1.4, 1.8, 1.6]):
        test = spdpotential._nonbond_potential(  # noqa: SLF001
            distance=distances,
            sigmas=np.array(_sigma),
        )
        assert np.argmin(test**2) == nb_mins[i]



import mchammer as mch  # noqa: INP001, D100
import spindry as spd
import stk

# Building a cage from the examples on the stk docs.
bb1 = stk.BuildingBlock(
    smiles="O=CC(C=O)C=O",
    functional_groups=[stk.AldehydeFactory()],
)
bb2 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])

cage = stk.ConstructedMolecule(
    topology_graph=stk.cage.FourPlusSix(
        building_blocks=(bb1, bb2),
    ),
)
stk_guests = (stk.BuildingBlock("C1CCCCC1"),)

host_guest = stk.ConstructedMolecule(
    topology_graph=stk.host_guest.Complex(
        host=stk.BuildingBlock.init_from_molecule(cage),
        guests=[stk.host_guest.Guest(i) for i in stk_guests],
    )
)

host_molecule = mch.Molecule(
    atoms=(
        mch.Atom(
            id=atom.get_id(),
            element_string=atom.__class__.__name__,
        )
        for atom in cage.get_atoms()
    ),
    bonds=(
        mch.Bond(
            id=i,
            atom_ids=(
                bond.get_atom1().get_id(),
                bond.get_atom2().get_id(),
            ),
        )
        for i, bond in enumerate(cage.get_bonds())
    ),
    position_matrix=(cage.get_position_matrix()),
)
guest_molecule = mch.Molecule(
    atoms=(
        mch.Atom(
            id=atom.get_id(),
            element_string=atom.__class__.__name__,
        )
        for atom in stk_guests[0].get_atoms()
    ),
    bonds=(
        mch.Bond(
            id=i,
            atom_ids=(
                bond.get_atom1().get_id(),
                bond.get_atom2().get_id(),
            ),
        )
        for i, bond in enumerate(stk_guests[0].get_bonds())
    ),
    position_matrix=stk_guests[0].get_position_matrix(),
)

supramolecule = spd.SupraMolecule.init_from_components(
    components=(host_molecule, guest_molecule),
)
cg = spd.Spinner(
    step_size=0.2,
    rotation_step_size=1.0,
    num_conformers=5,
)

for conformer in cg.get_conformers(supramolecule):  # noqa: B007
    pass

# Write optimised structures out.
for i, comp in enumerate(list(conformer.get_components())):
    if i == 0:
        new_cage = cage.with_position_matrix(comp.get_position_matrix())
        new_cage.write("comp_new_host.mol")
    elif i == 1:
        new_guest = stk_guests[0].with_position_matrix(
            comp.get_position_matrix()
        )
        new_guest.write("comp_new_guest.mol")



"""Classes for calculating the potential energy of supramolecules."""

from __future__ import annotations

import itertools as it
from typing import TYPE_CHECKING

import numpy as np
from scipy.spatial.distance import cdist

if TYPE_CHECKING:
    from .supramolecule import SupraMolecule


class Potential:
    """Base class for potential calculators."""

    def compute_potential(self, supramolecule: SupraMolecule) -> float:
        """Calculate potential energy.

        Parameters:
            supramolecular:
                Supramolecule to evaluate.
        """
        raise NotImplementedError


class SpdPotential(Potential):
    """Default spindry non-bonded potential function."""

    def __init__(self, nonbond_epsilon: float = 5) -> None:
        """Initialize a :class:`Spinner` instance.

        Parameters:
            nonbond_epsilon:
                Value of epsilon used in the nonbond potential in MC moves.
                Determines strength of the nonbond potential.
                Defaults to 5.

        """
        self._nonbond_epsilon = nonbond_epsilon

    def _nonbond_potential(
        self,
        distance: np.ndarray,
        sigmas: np.ndarray,
    ) -> np.ndarray:
        """Define a Lennard-Jones nonbonded potential.

        This potential has no relation to an empircal forcefield.

        """
        return self._nonbond_epsilon * (
            (sigmas / distance) ** 12 - (sigmas / distance) ** 6
        )

    def _combine_sigma(self, radii1: tuple, radii2: tuple) -> np.ndarray:
        """Combine radii using Lorentz-Berthelot rules."""
        len1 = len(radii1)
        len2 = len(radii2)

        mixed = np.zeros((len1, len2))
        for i in range(len1):
            for j in range(len2):
                mixed[i, j] = (radii1[i] + radii2[j]) / 2

        return mixed

    def _compute_nonbonded_potential(
        self,
        position_matrices: list[np.ndarray],
        radii: list[tuple],
    ) -> float:
        nonbonded_potential = 0
        for pos_mat_pair, radii_pair in zip(
            it.combinations(position_matrices, 2),
            it.combinations(radii, 2),
            strict=True,
        ):
            pair_dists = cdist(pos_mat_pair[0], pos_mat_pair[1])
            sigmas = self._combine_sigma(radii_pair[0], radii_pair[1])
            nonbonded_potential += np.sum(
                self._nonbond_potential(
                    distance=pair_dists.flatten(),
                    sigmas=sigmas.flatten(),
                )
            )

        return nonbonded_potential

    def compute_potential(self, supramolecule: SupraMolecule) -> float:
        """Compute the potential energy of a supramolecule."""
        component_position_matrices = [
            i.get_position_matrix() for i in supramolecule.get_components()
        ]
        component_radii = [
            tuple(j.get_radius() for j in i.get_atoms())
            for i in supramolecule.get_components()
        ]
        return self._compute_nonbonded_potential(
            position_matrices=component_position_matrices,
            radii=component_radii,
        )


class VaryingEpsilonPotential(Potential):
    """A non-bonded potential function with varying epsilons."""

    def _nonbond_potential(
        self,
        distance: np.ndarray,
        sigmas: np.ndarray,
        epsilons: np.ndarray,
    ) -> np.ndarray:
        """Define a Lennard-Jones nonbonded potential.

        This potential has no relation to an empircal forcefield.

        """
        return epsilons * (
            (sigmas / distance) ** 12 - (sigmas / distance) ** 6
        )

    def _combine_sigma(self, radii1: tuple, radii2: tuple) -> np.ndarray:
        """Combine radii using Lorentz-Berthelot rules."""
        len1 = len(radii1)
        len2 = len(radii2)

        mixed = np.zeros((len1, len2))
        for i in range(len1):
            for j in range(len2):
                mixed[i, j] = (radii1[i] + radii2[j]) / 2

        return mixed

    def _combine_epsilon(self, e1: tuple, e2: tuple) -> np.ndarray:
        """Combine epsilon using Lorentz-Berthelot rules."""
        len1 = len(e1)
        len2 = len(e2)

        mixed = np.zeros((len1, len2))
        for i in range(len1):
            for j in range(len2):
                mixed[i, j] = np.sqrt(e1[i] * e2[j])

        return mixed

    def _compute_nonbonded_potential(
        self,
        position_matrices: list[np.ndarray],
        radii: list[tuple],
        epsilons: list[tuple],
    ) -> float:
        nonbonded_potential = 0
        for pos_mat_pair, radii_pair, epsilon_pair in zip(
            it.combinations(position_matrices, 2),
            it.combinations(radii, 2),
            it.combinations(epsilons, 2),
            strict=True,
        ):
            pair_dists = cdist(pos_mat_pair[0], pos_mat_pair[1])
            new_radii = self._combine_sigma(radii_pair[0], radii_pair[1])
            new_epsilons = self._combine_epsilon(
                epsilon_pair[0], epsilon_pair[1]
            )
            nonbonded_potential += np.sum(
                self._nonbond_potential(
                    distance=pair_dists.flatten(),
                    sigmas=new_radii.flatten(),
                    epsilons=new_epsilons.flatten(),
                )
            )

        return nonbonded_potential

    def compute_potential(self, supramolecule: SupraMolecule) -> float:
        """Compure the potential of the molecule."""
        component_position_matrices = [
            i.get_position_matrix() for i in supramolecule.get_components()
        ]
        component_radii = [
            tuple(j.sigma for j in i.get_atoms())
            for i in supramolecule.get_components()
        ]
        component_epsilon = [
            tuple(j.epsilon for j in i.get_atoms())
            for i in supramolecule.get_components()
        ]
        return self._compute_nonbonded_potential(
            position_matrices=component_position_matrices,
            radii=component_radii,
            epsilons=component_epsilon,
        )



"""SpinDry package."""

from mchammer import Atom, Bond, Molecule

from spindry._internal.potential import (
    Potential,
    SpdPotential,
    VaryingEpsilonPotential,
)
from spindry._internal.spinner import Spinner
from spindry._internal.supramolecule import SupraMolecule
from spindry._internal.utilities import (
    calculate_centroid_distance,
    calculate_min_atom_distance,
    get_atom_distance,
)

__all__ = [
    "SpdPotential",
    "VaryingEpsilonPotential",
    "SupraMolecule",
    "Spinner",
    "Potential",
    "get_atom_distance",
    "calculate_min_atom_distance",
    "calculate_centroid_distance",
    "Atom",
    "Bond",
    "Molecule",
]
